from ibis_mssqlops.funcs import *
from ibis.expr import types as ir

# NOTE: Funcoes de texto
ir.StringValue.replicate = replicate
ir.StringValue.collatelatin = collatelatin
ir.StringValue.trim = trim

# NOTE: Funcoes para varios tipos
ir.Value.convert = convert
ir.Value.format = formats

# NOTE: Funcoes para True/False
ir.Value.isnumeric = isnumeric

# NOTE: Funcoes de data/hora
ir.DateValue.datefromparts = datefromparts
ir.TimestampValue.datefromparts = datefromparts
ir.DateValue.dateadd = dateadd
ir.TimestampValue.dateadd = dateadd
ir.DateValue.datediff = datediff
ir.TimestampValue.datediff = datediff
ir.DateValue.datepart = datepart
ir.TimestampValue.datepart = datepart


import ibis_mssqlops.addibis
import ibis_mssqlops.addsql
from ibis_mssqlops.consts import (
    DatePartType as datetype,
    MssqlTypeName as mssqltype,
    Style as style
)
from ibis_mssqlops.connect import (
    Connect as mssql_connect
)
import ibis
from ibis import deferred as _
from ibis import selectors as s
from ibis import udf
from ibis.expr import types as ir

ibis.options.interactive = True

__all__ = [
    "_", 
    "ibis", 
    "s", 
    "udf",
    "datetype", 
    "mssqltype",
    "style",
    "mssql_connect",
    "ir"
]

__version__ = '0.0.57'


import ibis
import sqlalchemy as sa
from ibis_mssqlops.expr import *


@ibis.mssql.add_operation(Format)
def _format(translator, expr):
    arg, format, culture = expr.args

    compile_arg = translator.translate(arg)
    compile_format = translator.translate(format)

    if culture is not None:
        compile_culture = translator.translate(culture)
        return sa.func.FORMAT(
            compile_arg,
            compile_format,
            compile_culture
        )
    
    return sa.func.FORMAT(
        compile_arg,
        compile_format
    )
    

@ibis.mssql.add_operation(DatePart)
def _datepart(translator, expr):
    arg, datepart = expr.args

    compile_arg = translator.translate(arg)

    return sa.func.DATEPART(
        sa.text(datepart.value),
        compile_arg
    )


@ibis.mssql.add_operation(DateDiff)
def _datediff(translator, expr):
    arg, enddate, datepart = expr.args

    compile_arg = translator.translate(arg)
    compile_enddate = translator.translate(enddate)

    return sa.func.DATEDIFF(
        sa.text(datepart.value),
        compile_arg,
        compile_enddate
    )


@ibis.mssql.add_operation(Replicate)
def _replicate(translator, expr):
    arg, count = expr.args

    compile_arg = translator.translate(arg)
    compile_count = translator.translate(count)

    return sa.func.REPLICATE(compile_arg, compile_count)


@ibis.mssql.add_operation(DateAdd)
def _dateadd(translator, expr):
    arg, datepart, number = expr.args

    compile_arg = translator.translate(arg)
    compile_number = translator.translate(number)

    return sa.func.DATEADD(
        sa.text(datepart.value),
        compile_number,
        compile_arg
    )


@ibis.mssql.add_operation(CollateLatin)
def _collatelatin(translator, expr):
    (arg, ) = expr.args

    compile_arg = translator.translate(arg)
    cast_comp = sa.cast(compile_arg, sa.VARCHAR(None))

    return sa.collate(cast_comp, 'SQL_Latin1_General_CP1251_CS_AS')


@ibis.mssql.add_operation(Trim)
def _trim(translator, expr):
    (arg, ) = expr.args

    compile_arg = translator.translate(arg)

    return sa.func.LTRIM(sa.func.RTRIM(compile_arg))


@ibis.mssql.add_operation(IsNumeric)
def _isnumeric(translator, expr):
    (arg, ) = expr.args

    compile_arg = translator.translate(arg) 
    
    return sa.func.ISNUMERIC(compile_arg)


@ibis.mssql.add_operation(DateFromParts)
def _datefromparts(translator, expr):
    arg, year, month, day = expr.args
    compile_arg = translator.translate(arg)

    def compila(param, op, arg=compile_arg):
        if param is not None:
            return translator.translate(param)
        
        match op:
            case 1:
                return sa.func.YEAR(arg)
            case 2:
                return sa.func.MONTH(arg)
            case 3:
                return sa.func.DAY(arg)

    year_comp = compila(year, 1)
    month_comp = compila(month, 2)
    day_comp = compila(day, 3)

    return sa.func.DATEFROMPARTS(
        year_comp, 
        month_comp, 
        day_comp
    )


@ibis.mssql.add_operation(Convert)
def _convert(translator, expr):
    arg, data_type, style = expr.args

    compile_arg = translator.translate(arg)

    if style is not None:
        compile_style = translator.translate(style)
        return sa.func.CONVERT(
            sa.text(data_type.value), 
            compile_arg, 
            compile_style
        )

    return sa.func.CONVERT(
        sa.text(data_type.value), 
        compile_arg
    )


from ibis_mssqlops.expr import *
from ibis_mssqlops.transforms import *


def formats(
    value,
    format,
    culture = None
):
    
    return Format(
        value,
        format=format,
        culture=culture
    ).to_expr()


@get_value_enums
def datepart(
    expression,
    datepart
):
    
    return DatePart(
        expression,
        datepart=datepart
    ).to_expr()


@get_value_enums
def datediff(
    startdate,
    enddate,
    datepart = 'month'
):
    
    return DateDiff(
        startdate,
        enddate=enddate,
        datepart=datepart
    ).to_expr()


def replicate(
    expression,
    count
):
    return Replicate(
        expression,
        count=count
    ).to_expr()


@get_value_enums
def dateadd(
    expression, 
    datepart, 
    number
):

    return DateAdd(
        expression, 
        datepart=datepart, 
        number=number
    ).to_expr()


def collatelatin(expression):
    return CollateLatin(expression).to_expr()


def trim(expression):
    return Trim(expression).to_expr()


def isnumeric(expression):
    return IsNumeric(expression).to_expr()


def datefromparts(
    expression, 
    year = None, 
    month = None, 
    day = None
):
    return DateFromParts(
        expression, 
        year=year, 
        month=month, 
        day=day
    ).to_expr()


@get_value_enums
def convert(
    expression, 
    data_type, 
    style = None
):

    return Convert(
        expression, 
        data_type=data_type, 
        style=style
    ).to_expr()


from ibis.expr import datatypes as dt


_from_type_dtype = {
    'BIGINT': dt.int64,
    'BIT': dt.boolean,
    'DECIMAL': dt.decimal,
    'INT': dt.int32,
    'MONEY': dt.int64,
    'NUMERIC': dt.decimal,
    'SMALLINT': dt.int16,
    'SMALLMONEY': dt.int32,
    'TINYINT': dt.int8,
    'FLOAT': dt.float64,
    'REAL': dt.float32,
    'DATE': dt.date,
    'DATETIME2': dt.timestamp,
    'DATETIME': dt.timestamp,
    'DATETIMEOFFSET': dt.timestamp,
    'SMALLDATETIME': dt.timestamp,
    'TIME': dt.time,
    'CHAR': dt.string,
    'TEXT': dt.string,
    'VARCHAR': dt.string,
    'NCHAR': dt.string,
    'NTEXT': dt.string,
    'NVARCHAR': dt.string,
    'BINARY': dt.binary,
    'IMAGE': dt.binary,
    'VARBINARY': dt.binary,
    'UNIQUEIDENTIFIER': dt.uuid,
    'GEOMETRY': dt.GeoSpatial,
    'GEOGRAPHY': dt.GeoSpatial,
    'TIMESTAMP': dt.binary
}

__all__ = ['_from_type_dtype']


from ibis_mssqlops.consts import (
    DatePartType,
    MssqlTypeName,
    Style
)
from functools import wraps


def get_value_enums(func):
    @wraps(func)
    def inner(*args, **kwargs):
        tps = (
            Style,
            DatePartType,
            MssqlTypeName
        )

        is_enum = lambda x: isinstance(x, tps)

        args_alter = tuple(
            arg.value if is_enum(arg) 
            else arg 
            for arg in args
        )

        kwargs_alter = {
            k:v.value if is_enum(v) 
            else v 
            for k, v in kwargs.items()
        }

        return func(*args_alter, **kwargs_alter)
    return inner


from ibis.expr import (
    datatypes as dt,
    datashape as ds,
    rules as rlz
)
from ibis.expr.operations import Value
from ibis_mssqlops.datatypes import _from_type_dtype


class Format(Value):
    arg: Value[dt.Any, ds.Any]
    format: Value[dt.String, ds.Any]
    culture: Value[dt.String, ds.Any] | None = None

    dtype = dt.string
    shape = rlz.shape_like('arg')


class DatePart(Value):
    arg: Value[dt.Timestamp | dt.Date, ds.Any]
    datepart: Value[dt.String, ds.Any]

    dtype = dt.int
    shape = rlz.shape_like('arg')


class DateDiff(Value):
    arg: Value[dt.Timestamp | dt.Date, ds.Any]
    enddate: Value[dt.Timestamp | dt.Date, ds.Any]
    datepart: Value[dt.String, ds.Any]

    dtype = dt.int
    shape = rlz.shape_like('arg')


class Replicate(Value):
    arg: Value[dt.String, ds.Any]
    count: Value[dt.Integer, ds.Any]

    dtype = dt.string
    shape = rlz.shape_like('arg')
    

class DateAdd(Value):
    arg: Value[dt.Date | dt.Timestamp, ds.Any]
    datepart: Value[dt.String, ds.Any]
    number: Value[dt.Integer, ds.Any]
    
    dtype = rlz.dtype_like('arg')
    shape = rlz.shape_like('arg')


class CollateLatin(Value):
    arg: Value[dt.String, ds.Any]

    dtype = dt.string
    shape = rlz.shape_like('arg')


class Trim(Value):
    arg: Value[dt.String, ds.Any]

    dtype = dt.string
    shape = rlz.shape_like('arg')


class IsNumeric(Value):
    arg: Value[dt.Any, ds.Any]

    dtype = dt.bool
    shape = rlz.shape_like('arg')


class Convert(Value):
    arg: Value[dt.Any, ds.Any]
    
    data_type: Value[dt.String, ds.Any]
    style: Value[dt.Integer, ds.Any] | None = None

    shape = rlz.shape_like('arg')

    @property
    def dtype(self) -> dt.DataType:
        key = self.data_type.value
        return _from_type_dtype[key.upper()]
    

class DateFromParts(Value):
    arg: Value[dt.Date | dt.Timestamp, ds.Any]
    
    year: Value[dt.Integer, ds.Any] | None = None
    month: Value[dt.Integer, ds.Any] | None = None
    day: Value[dt.Integer, ds.Any] | None = None

    dtype = dt.date
    shape = rlz.shape_like('arg')


from enum import Enum


_dtype = {
    1: 'mm/dd/yy',
    101: 'mm/dd/yyyy',
    2: 'yy.mm.dd',
    102: 'yyyy.mm.dd',
    3: 'dd/mm/yy',
    103: 'dd/mm/yyyy',
    4: 'dd.mm.yy',
    104: 'dd.mm.yyyy',
    5: 'dd-mm-yyy',
    105: 'dd-mm-yyyy'
}


class Style(Enum):
    MMDDYY = 1
    MMDDYYYY = 101
    YYMMDD = 2
    YYYYMMDD = 102
    DDMMYY = 3
    DDMMYYYY = 103
    DD_MM_YY = 4
    DD_MM_YYYY = 104
    DD__MM__YY = 5
    DD__MM__YYYY = 105

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f'{cls_name}.{_dtype[self.value]}'
    
    def __str__(self) -> str:
        return f'{_dtype[self.value]}'


class MssqlTypeName(str, Enum):
    BIGINT = 'BIGINT'
    BIT = 'BIT'
    DECIMAL = 'DECIMAL'
    INT = 'INT'
    MONEY = 'MONEY'
    NUMERIC = 'NUMERIC'
    SMALLINT = 'SMALLINT'
    SMALLMONEY = 'SMALLMONEY'
    TINYINT = 'TINYINT'
    FLOAT = 'FLOAT'
    REAL = 'REAL'
    DATE = 'DATE'
    DATETIME2 = 'DATETIME2'
    DATETIME = 'DATETIME'
    DATETIMEOFFSET = 'DATETIMEOFFSET'
    SMALLDATETIME = 'SMALLDATETIME'
    TIME = 'TIME'
    CHAR = 'CHAR'
    TEXT = 'TEXT'
    VARCHAR = 'VARCHAR'
    NCHAR = 'NCHAR'
    NTEXT = 'NTEXT'
    NVARCHAR = 'NVARCHAR'
    BINARY = 'BINARY'
    IMAGE = 'IMAGE'
    VARBINARY = 'VARBINARY'
    UNIQUEIDENTIFIER = 'UNIQUEIDENTIFIER'
    GEOMETRY = 'GEOMETRY'
    GEOGRAPHY = 'GEOGRAPHY'
    TIMESTAMP = 'TIMESTAMP'


class DatePartType(str, Enum):
    YEAR = 'year' 
    QUARTER = 'quarter'
    MONTH = 'month'
    DAYOFYEAR = 'dayofyear' 
    DAY = 'day'
    WEEK =	'week'
    WEEKDAY = 'weekday'
    HOUR = 'hour'
    MINUTE = 'minute'
    SECOND = 'second'
    MILLISECOND = 'millisecond' 
    MICROSECOND = 'microsecond'
    NANOSECOND = 'nanosecond'


__all__ = [
    'Style',
    'MssqlTypeName',
    'DatePartType'
]


import ibis
import ibis.expr.types as ir
from typing import (
    Sequence,
    Literal
)
import pandas as pd
import pyarrow as pa
from functools import wraps


class Connect:
    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs
        self.con = None

    def __connect(self) -> ibis.BaseBackend:
        return (
            ibis.mssql.connect(
                *self.args,
                **self.kwargs
            )
        )
    
    def is_connect(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if self.con is None:
                self.con = self.__connect()
            return func(self, *args, **kwargs)
        return inner
    
    @is_connect
    def list_databases(
        self, 
        *args, 
        **kwargs
    ) -> list[str]:

        return self.con.list_databases(*args, **kwargs)
    
    @is_connect
    def list_tables(
        self, 
        *args, 
        **kwargs
    ) -> list[str]:
        
        return self.con.list_tables(*args, **kwargs)
    
    @is_connect
    def table(
        self,
        tbl: str, 
        /,
        *,
        columns: Sequence[str] = None,
        server: str = None,
        database: str = None,
        schema: str = 'dbo',
        key: Literal['nolock', 'readpast'] = 'nolock'
    ) -> ir.Table:
        
        cols = (
            ','.join(columns)
            if columns
            else '*'
        )

        tree = ''.join(
            [
               f'{server}.' if server else '',
               f'{database}.' if database else '',
               f'{schema}.',
               tbl
            ]
        )

        sel = f'''
           SELECT {cols} 
           FROM {tree} 
           WITH({key})
        '''

        return (
            self.con.sql(sel)
        )
    
    def __insert_pandas(
        self, 
        tbl: str,
        df: pd.DataFrame
    ) -> ir.Table:
        
        schema = ibis.memtable(df).schema()

        self.con.drop_table(tbl, force=True)
        temp = self.con.create_table(
            tbl, 
            schema=schema, 
            overwrite=True
        )

        self.con.insert(tbl, df)

        return temp
    
    @is_connect
    def create_temp_table(
        self,
        tbl: str,
        obj: pd.DataFrame | ir.Table | pa.Table,
        is_global: bool = False,
        **kwargs
    ) -> ir.Table:
        
        prefix = '#'
        if is_global:
            prefix += '#'

        if isinstance(obj, pd.DataFrame):
            return self.__insert_pandas(f'{prefix}{tbl}', obj)

        return self.con.create_table(f'{prefix}{tbl}', obj, **kwargs)
    
    @is_connect
    def drop_temp_table(
        self,
        tbl,
        is_global: bool = False,
        **kwargs
    ) -> None:
        
        prefix = '#'
        if is_global:
            prefix += '#'
        
        return self.con.drop_table(f'{prefix}{tbl}', **kwargs)